/*     
 **********************************************************************
 *     efxmgr.c
 *     Copyright 1999, 2000 Creative Labs, Inc. 
 * 
 ********************************************************************** 
 * 
 *     Date                 Author          Summary of changes 
 *     ----                 ------          ------------------ 
 *     October 20, 1999     Bertrand Lee    base code release 
 * 
 ********************************************************************** 
 * 
 *     This program is free software; you can redistribute it and/or 
 *     modify it under the terms of the GNU General Public License as 
 *     published by the Free Software Foundation; either version 2 of 
 *     the License, or (at your option) any later version. 
 * 
 *     This program is distributed in the hope that it will be useful, 
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of 
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 *     GNU General Public License for more details. 
 * 
 *     You should have received a copy of the GNU General Public 
 *     License along with this program; if not, write to the Free 
 *     Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, 
 *     USA. 
 * 
 ********************************************************************** 
 */

#include <linux/bitops.h>
#include "hwaccess.h"
#include "efxmgr.h"

int emu10k1_find_control_gpr(struct patch_manager *mgr, const char *patch_name, const char *gpr_name)
{
        struct dsp_patch *patch;
	struct dsp_rpatch *rpatch;
	char s[PATCH_NAME_SIZE + 4];
	u32 *gpr_used;
	int i;

	DPD(2, "emu10k1_find_control_gpr(): %s %s\n", patch_name, gpr_name);

	rpatch = &mgr->rpatch;
	if (!strcmp(rpatch->name, patch_name)) {
		gpr_used = rpatch->gpr_used;
		goto match;
	}

	for(i = 0; i < mgr->current_pages * PATCHES_PER_PAGE; i++) {
		patch = PATCH(mgr, i);
			sprintf(s,"%s", patch->name);

		if (!strcmp(s, patch_name)) {
			gpr_used = patch->gpr_used;
			goto match;
		}
	}

	return -1;

  match:
	for (i = 0; i < NUM_GPRS; i++)
		if (mgr->gpr[i].type == GPR_TYPE_CONTROL &&
		    test_bit(i, gpr_used) &&
		    !strcmp(mgr->gpr[i].name, gpr_name))
			return i;

	return -1;
}

void emu10k1_set_control_gpr(struct emu10k1_card *card, int addr, s32 val, int flag)
{
	struct patch_manager *mgr = &card->mgr;

	DPD(2, "emu10k1_set_control_gpr(): %d %x\n", addr, val);

	if (addr < 0 || addr >= NUM_GPRS)
		return;

	if (flag)
		val += sblive_readptr(card, GPR_BASE + addr, 0);

	if (val > mgr->gpr[addr].max)
		val = mgr->gpr[addr].max;
	else if (val < mgr->gpr[addr].min)
		val = mgr->gpr[addr].min;

	sblive_writeptr(card, GPR_BASE + addr, 0, val);
}

#define VOL_CTRL_PATCH_NAME     "hw vol ctrl"
#define VOL_CTRL_GPR_NAME       "Vol_ctrl"
#define STEP_SIZE       0x08000000

void emu10k1_mute_irqhandler(struct emu10k1_card *card)
{
	struct patch_manager *mgr = &card->mgr;
	int addr;
	unsigned long flags;

	DPF(2, "Vol_ctrl mute button pressed\n");

	spin_lock_irqsave(&mgr->lock, flags);
	addr = emu10k1_find_control_gpr(mgr, VOL_CTRL_PATCH_NAME, VOL_CTRL_GPR_NAME);
	emu10k1_set_control_gpr(card, addr, 0x0, 0);
	spin_unlock_irqrestore(&mgr->lock, flags);
}

void emu10k1_volincr_irqhandler(struct emu10k1_card *card)
{
	struct patch_manager *mgr = &card->mgr;
	int addr;
	unsigned long flags;

	DPF(2, "Vol_incr button pressed\n");

	spin_lock_irqsave(&mgr->lock, flags);
	addr = emu10k1_find_control_gpr(mgr, VOL_CTRL_PATCH_NAME, VOL_CTRL_GPR_NAME);
	emu10k1_set_control_gpr(card, addr, STEP_SIZE, 1);
	spin_unlock_irqrestore(&mgr->lock, flags);
}

void emu10k1_voldecr_irqhandler(struct emu10k1_card *card)
{
	struct patch_manager *mgr = &card->mgr;
	int addr;
	unsigned long flags;

	DPF(2, "Vol_decr button pressed\n");
	spin_lock_irqsave(&mgr->lock, flags);
	addr = emu10k1_find_control_gpr(mgr, VOL_CTRL_PATCH_NAME, VOL_CTRL_GPR_NAME);
	emu10k1_set_control_gpr(card, addr, -STEP_SIZE, 1);
	spin_unlock_irqrestore(&mgr->lock, flags);
}


void emu10k1_dsp_volume(struct emu10k1_card *card, const char *patch_name, const char *gpr_name, s32 vol, int scale, int muting)
{
	struct patch_manager *mgr = &card->mgr;
	int addr;
	unsigned long flags;

	const s32 log2lin[5] ={                  //  attenuation (dB)
		0x7fffffff,                      //       0.0         
		0x7fffffff * 0.840896415253715 , //       1.5          
		0x7fffffff * 0.707106781186548,  //       3.0
		0x7fffffff * 0.594603557501361 , //       4.5
	};

	vol = (100 - vol ) * scale / 100;

	// Thanks to the comp.dsp newsgroup for this neat trick:
	vol = vol >= muting ? 0: log2lin[vol&3]>>(vol>>2);

	spin_lock_irqsave(&mgr->lock, flags);
	addr = emu10k1_find_control_gpr(mgr, patch_name, gpr_name);
	emu10k1_set_control_gpr(card, addr, vol, 0);
	spin_unlock_irqrestore(&mgr->lock, flags);
}

void emu10k1_dsp_irqhandler(struct emu10k1_card *card)
{


}
