/*********************************************************************
 *     dsp.c - dsp patch manager library
 * 
 *      Copyright (C) 2000 Rui Sousa 
 ********************************************************************* 
 *     This program is free software; you can redistribute it and/or 
 *     modify it under the terms of the GNU General Public License as 
 *     published by the Free Software Foundation; either version 2 of 
 *     the License, or (at your option) any later version. 
 * 
 *     This program is distributed in the hope that it will be useful, 
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of 
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 *     GNU General Public License for more details. 
 * 
 *     You should have received a copy of the GNU General Public 
 *     License along with this program; if not, write to the Free 
 *     Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, 
 *     USA. 
 ********************************************************************* 
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <fcntl.h>

#include "../include/dsp.h"
#include "../include/list.h"

#define MIXER_DEV_NAME "/dev/mixer"
#define DSP_DEV_NAME "/dev/dsp"

#define OUTPUT_BASE 0x20
#define CONSTANTS_BASE 0x40
#define GPR_BASE 0x100
#define TRAML_IDATA_BASE 0x200
#define TRAML_EDATA_BASE 0x280
#define TRAML_IADDR_BASE 0x300
#define TRAML_EADDR_BASE 0x380
#define DSP_CODE_BASE 0x400

#define CONSTANTS_SIZE 0x16
#define GPR_SIZE 0x100
#define TRAML_ISIZE 0x80
#define TRAML_ESIZE 0x20
#define TRAMB_ISIZE 0x2000
#define TRAMB_ESIZE 0x200000
#define DSP_CODE_SIZE 0x400

__u32 addr_change[0x400];

__s32 constants[CONSTANTS_SIZE] = {
	0x0,
	0x1,
	0x2,
	0x3,
	0x4,
	0x8,
	0x10,
	0x20,
	0x100,
	0x10000,
	0x80000,
	0x10000000,
	0x20000000,
	0x40000000,
	0x80000000,
	0x7fffffff,
	0xffffffff,
	0xfffffffe,
	0xc0000000,
	0x4f1bbcdc,
	0x5a7ef9db,
	0x00100000
};

/* FIXME */
/* these should have better names (and probably dependent on the card model) */
char *dsp_in_name[DSP_NUM_INPUTS] = { "Pcm L", "Pcm R", "Pcm1 L", "Pcm1 R", "fx4", "fx5", "fx6", "fx7", "fx8",
	"fx9", "fx10", "fx11", "fx12", "fx13", "fx14", "fx15",
	"Analog L", "Analog R", "CD-Spdif L", "CD-Spdif R", "in2l", "in2r", "Opt. Spdif L", "Opt. Spdif R",
	"RCA Aux L", "RCA Aux R", "RCA Spdif L", "RCA Spdif R", "Line2/Mic2 L", "Line2/Mic2 R", "in7l", "in7r"
};

char *dsp_out_name[DSP_NUM_OUTPUTS] =
    { "Front L", "Front R", "Digital L", "Digital R", "out2l", "out2r", "Phones L", "Phones R",
	"Rear L", "Rear R", "ADC Rec L", "ADC Rec R", "Mic Rec", "out6r", "out7l", "out7r",
	"out8l", "out8r", "out9l", "out9r", "out10l", "out10r", "out11l", "out11r",
	"out12l", "out12r", "out13l", "out13r", "out14l", "out14r", "out15l", "out15r"
};

struct kdsp_patch {
	char name[DSP_PATCH_NAME_SIZE];
	__u8 id;
	__u32 line;		/* bitmap of the lines the patch is attached to */
	__u16 code_start;
	__u16 code_size;

	__u32 gpr_used[DSP_NUM_GPRS / 32];	/* bitmap of used gprs */
	__u32 gpr_input[DSP_NUM_GPRS / 32];
	__u8 traml_istart;	/* starting address of the internal tram lines used */
	__u8 traml_isize;	/* number of internal tram lines used */

	__u8 traml_estart;
	__u8 traml_esize;

	__u16 tramb_istart;	/* starting address of the internal tram memory used */
	__u16 tramb_isize;	/* amount of internal memory used */
	__u32 tramb_estart;
	__u32 tramb_esize;
};

struct dsp_kernel_rpatch {
	char name[DSP_PATCH_NAME_SIZE];
	__u16 code_start;
	__u16 code_size;

	__u32 gpr_used[DSP_NUM_GPRS / 32];
	__u32 gpr_input[DSP_NUM_GPRS / 32];
	__u32 route[DSP_NUM_OUTPUTS];
	__u32 route_v[DSP_NUM_OUTPUTS];
};

enum {
	GPR_TYPE_NULL = 0,
	GPR_TYPE_IO,
	GPR_TYPE_STATIC,
	GPR_TYPE_DYNAMIC,
	GPR_TYPE_CONTROL,
	GPR_TYPE_CONSTANT
};

#define TRAML_TYPE_READ 0x1
#define TRAML_TYPE_WRITE 0x2

#define CMD_READ 1
#define CMD_WRITE 2

struct mixer_private_ioctl {
	__u32 cmd;
	__u32 val[90];		/* in kernel space only a fraction of this is used */
};

#define CMD_WRITEPTR		_IOW('D', 2, struct mixer_private_ioctl)
#define CMD_READPTR		_IOR('D', 3, struct mixer_private_ioctl)
#define CMD_GETPATCH		_IOR('D', 8, struct mixer_private_ioctl)
#define CMD_GETGPR		_IOR('D', 9, struct mixer_private_ioctl)
#define CMD_GETCTLGPR		_IOR('D', 10, struct mixer_private_ioctl)
#define CMD_SETPATCH		_IOW('D', 11, struct mixer_private_ioctl)
#define CMD_SETGPR		_IOW('D', 12, struct mixer_private_ioctl)
#define CMD_SETCTLGPR		_IOW('D', 13, struct mixer_private_ioctl)

#define PATCH_INPUT	0x1
#define PATCH_OUTPUT	0x2
#define GPR_INPUT	0x3
#define GPR_OUTPUT	0x4

static void determine_io(struct dsp_patch_manager *, struct dsp_patch *);

static int find_control_gpr_addr(struct dsp_patch_manager *mgr, const char *patch_name, const char *gpr_name)
{
	struct dsp_rpatch *rpatch;
	struct dsp_patch *patch;
	struct list_head *entry;
	char s[DSP_PATCH_NAME_SIZE + 4];
	__u32 *gpr_used;
	int i;

	for (i = 0; i < DSP_NUM_INPUTS; i++)
		list_for_each(entry, &mgr->ipatch_list[i]) {
		patch = list_entry(entry, struct dsp_patch, list);
		if (patch->id)
			sprintf(s, "%s %d", patch->name, patch->id);
		else
			sprintf(s, "%s", patch->name);

		if (!strcmp(s, patch_name)) {
			gpr_used = patch->gpr_used;
			goto match;
		}
		}

	rpatch = &mgr->rpatch;
	if (!strcmp(rpatch->name, patch_name)) {
		gpr_used = rpatch->gpr_used;
		goto match;
	}

	for (i = 0; i < DSP_NUM_OUTPUTS; i++)
		list_for_each(entry, &mgr->opatch_list[i]) {
		patch = list_entry(entry, struct dsp_patch, list);
		if (patch->id)
			sprintf(s, "%s %d", patch->name, patch->id);
		else
			sprintf(s, "%s", patch->name);

		if (!strcmp(s, patch_name)) {
			gpr_used = patch->gpr_used;
			goto match;
		}
		}

	return -1;

      match:

	for (i = 0; i < DSP_NUM_GPRS; i++) {
		if ((mgr->gpr[i].type == GPR_TYPE_CONTROL)
		    && test_bit(i, gpr_used)
		    && !strcmp(mgr->gpr[i].name, gpr_name))
			return GPR_BASE + i;
	}

	return -1;
}

int get_control_gpr_info(struct dsp_patch_manager *mgr, const char *patch_name, const char *gpr_name,
			 struct dsp_gpr *gpr)
{
	int addr;

	addr = find_control_gpr_addr(mgr, patch_name, gpr_name);
	if (addr < 0)
		return -1;

	memcpy(gpr, &mgr->gpr[addr - GPR_BASE], sizeof(struct dsp_gpr));

	return 0;
}

static void print_patch_control_gprs(struct dsp_patch_manager *mgr, __u32 * used)
{
	int i;

	for (i = 0; i < DSP_NUM_GPRS; i++)
		if (mgr->gpr[i].type == GPR_TYPE_CONTROL && test_bit(i, used)) {
			printf("    name: %s\n", mgr->gpr[i].name);
			printf("    addr: %#04x\n", mgr->gpr[i].addr);
			printf("    value:  %#010x (%#010x - %#010x)\n\n", mgr->gpr[i].value,
			       mgr->gpr[i].min_val, mgr->gpr[i].max_val);
		}
}

void dsp_print_control_gpr_list(struct dsp_patch_manager *mgr)
{
	struct dsp_rpatch *rpatch;
	struct dsp_patch *patch;
	struct list_head *entry;
	int i;

	if (!mgr->init)
		dsp_init(mgr);

	printf("Control GPR's:\n");

	for (i = 0; i < DSP_NUM_INPUTS; i++)
		list_for_each(entry, &mgr->ipatch_list[i]) {
		patch = list_entry(entry, struct dsp_patch, list);
		if (patch->id)
			printf("  Patch: %s %d\n", patch->name, patch->id);
		else
			printf("  Patch: %s\n", patch->name);

		print_patch_control_gprs(mgr, patch->gpr_used);
		}

	rpatch = &mgr->rpatch;
	printf("  Patch: %s\n", rpatch->name);
	print_patch_control_gprs(mgr, rpatch->gpr_used);

	for (i = 0; i < DSP_NUM_OUTPUTS; i++)
		list_for_each(entry, &mgr->opatch_list[i]) {
		patch = list_entry(entry, struct dsp_patch, list);
		if (patch->id)
			printf("  Patch: %s %d\n", patch->name, patch->id);
		else
			printf("  Patch: %s\n", patch->name);

		print_patch_control_gprs(mgr, patch->gpr_used);
		}
}

void dsp_print_route_list(struct dsp_patch_manager *mgr)
{
	struct dsp_rpatch *rpatch;
	int i, j;

	if (!mgr->init)
		dsp_init(mgr);

	rpatch = &mgr->rpatch;

	printf("\n  Routes:\n");

	for (j = 0; j < DSP_NUM_INPUTS; j++)
		for (i = 0; i < DSP_NUM_OUTPUTS; i++)
			if (test_bit(j, &rpatch->route[i]))
				printf("    %s:%s\n", dsp_in_name[j], dsp_out_name[i]);
			else if (test_bit(j, &rpatch->route_v[i]))
				printf("    %s:%s  (with volume control)\n", dsp_in_name[j], dsp_out_name[i]);

}

static void print_patch_code(__u32 * code, __u32 start, __u32 size)
{
	int i;

	printf("\n  code:\n");
	printf("    start: %#05x\n    size: %#05x\n", start, size);
	for (i = 0; i < size / 2; i++) {
		printf
		    ("    %#05x   %#010x  %#010x %2d %#05x %#05x %#05x %#05x\n",
		     DSP_CODE_BASE + start + 2 * i, code[i * 2 + 1],
		     code[i * 2], (code[i * 2 + 1] >> 20) & 0xf,
		     (code[i * 2 + 1] >> 10) & 0x3ff, code[i * 2 + 1] & 0x3ff,
		     (code[i * 2] >> 10) & 0x3ff, code[i * 2] & 0x3ff);
	}
}

static void print_patch_gprs(struct dsp_patch_manager *mgr, __u32 * used, __u32 * input)
{
	int i;

	printf("\n  gprs:");
	for (i = 0; i < DSP_NUM_GPRS; i++)
		if (test_bit(i, used)) {
			printf("\n");
			printf("    addr: %#05x\n", GPR_BASE + i);
			printf("    type: %#04x\n", mgr->gpr[i].type);

			if (mgr->gpr[i].type == GPR_TYPE_CONTROL) {
				printf("    name: %s\n", mgr->gpr[i].name);
				printf("    range: %#010x - %#010x\n", mgr->gpr[i].min_val, mgr->gpr[i].max_val);
			}

			if (mgr->gpr[i].type == GPR_TYPE_IO) {
				printf("    line: %d ", mgr->gpr[i].line);
				if (test_bit(i, input))
					printf("(Input)\n");
				else
					printf("(Output)\n");
			}

			if (mgr->gpr[i].type == GPR_TYPE_CONTROL ||
			    mgr->gpr[i].type == GPR_TYPE_STATIC || mgr->gpr[i].type == GPR_TYPE_CONSTANT)
				printf("    value: %#010x\n", mgr->gpr[i].value);

			printf("    usage: %#04x\n", mgr->gpr[i].usage);
		}
}

static void print_patch_tram(struct dsp_patch_manager *mgr, struct dsp_patch *patch)
{
	int i;

	if (patch->tramb_isize || patch->tramb_esize || patch->traml_isize || patch->traml_esize)
		printf("\n  tram:\n");

	if (patch->tramb_isize || patch->tramb_esize)
		printf("    blocks:\n");

	if (patch->tramb_isize)
		printf("      internal: %#07x - %#07x\n", patch->tramb_istart,
		       patch->tramb_istart + patch->tramb_isize);

	if (patch->tramb_esize)
		printf("      external: %#07x - %#07x\n", patch->tramb_estart,
		       patch->tramb_estart + patch->tramb_esize);

	if (patch->traml_isize || patch->traml_esize)
		printf("    lines:\n");

	if (patch->traml_isize) {
		printf("      data internal: %#05x - %#05x\n",
		       TRAML_IDATA_BASE + patch->traml_istart,
		       TRAML_IDATA_BASE + patch->traml_istart + patch->traml_isize - 1);

		printf("      address internal: %#05x - %#05x\n",
		       TRAML_IADDR_BASE + patch->traml_istart,
		       TRAML_IADDR_BASE + patch->traml_istart + patch->traml_isize - 1);

		for (i = 0; i < patch->traml_isize; i++)
			printf("        %#05x %#06x type: %d\n",
			       TRAML_IADDR_BASE + patch->traml_istart + i,
			       patch->traml_iaddr[i] & 0xfffff, patch->traml_iaddr[i] >> 20);
	}

	if (patch->traml_esize) {
		printf("      data external: %#05x - %#05x\n",
		       TRAML_EDATA_BASE + patch->traml_estart,
		       TRAML_EDATA_BASE + patch->traml_estart + patch->traml_esize - 1);

		printf("      address external: %#05x - %#05x\n",
		       TRAML_EADDR_BASE + patch->traml_estart,
		       TRAML_EADDR_BASE + patch->traml_estart + patch->traml_esize - 1);

		for (i = 0; i < patch->traml_esize; i++)
			printf("        %#05x %#07x type: %d\n",
			       TRAML_EADDR_BASE + patch->traml_estart + i,
			       patch->traml_eaddr[i] & 0xfffff, patch->traml_eaddr[i] >> 20);
	}
}

static void print_patch(struct dsp_patch_manager *mgr, struct dsp_patch *patch)
{
	printf("\n  Name: %s", patch->name);
	if (patch->id)
		printf(" %d", patch->id);
	printf("\n");
	print_patch_gprs(mgr, patch->gpr_used, patch->gpr_input);
	print_patch_code(patch->code, patch->code_start, patch->code_size);
	print_patch_tram(mgr, patch);
}

static void print_debug(struct dsp_patch_manager *mgr)
{
	struct dsp_rpatch *rpatch;
	struct dsp_patch *patch;
	struct list_head *entry;
	int i;

	printf("\nInput patches:\n");
	for (i = 0; i < DSP_NUM_INPUTS; i++)
		list_for_each(entry, &mgr->ipatch_list[i]) {
		patch = list_entry(entry, struct dsp_patch, list);
		print_patch(mgr, patch);
		}

	rpatch = &mgr->rpatch;

	printf("\nRouting patch:\n");

	print_patch_gprs(mgr, rpatch->gpr_used, rpatch->gpr_input);
	dsp_print_route_list(mgr);
	print_patch_code(rpatch->code, rpatch->code_start, rpatch->code_size);

	printf("\nOutput patches:\n");
	for (i = 0; i < DSP_NUM_OUTPUTS; i++)
		list_for_each(entry, &mgr->opatch_list[i]) {
		patch = list_entry(entry, struct dsp_patch, list);
		print_patch(mgr, patch);
		}
}

void dsp_print_patch_list(struct dsp_patch_manager *mgr)
{
	struct dsp_patch *patch;
	struct list_head *entry;
	int i;

	if (!mgr->init)
		dsp_init(mgr);

	for (i = 0; i < DSP_NUM_INPUTS; i++) {
		if (!list_empty(&mgr->ipatch_list[i])) {
			printf("%s", dsp_in_name[i]);
			list_for_each(entry, &mgr->ipatch_list[i]) {
				patch = list_entry(entry, struct dsp_patch, list);

				printf(" - %s", patch->name);
				if (patch->id)
					printf(" %d", patch->id);
			}

			printf("\n");
		}
	}
	for (i = 0; i < DSP_NUM_OUTPUTS; i++) {
		if (!list_empty(&mgr->opatch_list[i])) {
			list_for_each(entry, &mgr->opatch_list[i]) {
				patch = list_entry(entry, struct dsp_patch, list);
				printf("%s", patch->name);
				if (patch->id)
					printf(" %d", patch->id);
				printf(" - ");
			}

			printf("%s\n", dsp_out_name[i]);
		}
	}
}

static int get_constant_gpr(struct dsp_patch_manager *mgr, __s32 val, __u32 * used)
{
	int i;

	for (i = 0; i < CONSTANTS_SIZE; i++)
		/* if it's a hardwired constant use it */
		if (constants[i] == val)
			return CONSTANTS_BASE + i;

	for (i = 0; i < DSP_NUM_GPRS; i++) {
		/* see if he have this global gpr already */
		if (mgr->gpr[i].type == GPR_TYPE_CONSTANT && mgr->gpr[i].value == val && mgr->gpr[i].usage < 255)
			goto match;
	}

	for (i = 0; i < DSP_NUM_GPRS; i++) {
		/* assign a new free GPR */
		if (mgr->gpr[i].type == GPR_TYPE_NULL)
			goto match;
	}

	fprintf(stderr, "get_constant_gpr(): no available gprs\n");
	return -1;

      match:

	if (!test_and_set_bit(i, used)) {
		mgr->gpr[i].type = GPR_TYPE_CONSTANT;
		mgr->gpr[i].addr = i;
		mgr->gpr[i].usage++;
		mgr->gpr[i].value = val;
	}

	return GPR_BASE + i;
}

static int get_dynamic_gpr(struct dsp_patch_manager *mgr, __u32 * used)
{
	int i;

	for (i = 0; i < DSP_NUM_GPRS; i++) {
		/* overlap DYNAMIC gprs from different patches */
		if ((mgr->gpr[i].type == GPR_TYPE_DYNAMIC)
		    && !test_bit(i, used) && mgr->gpr[i].usage < 255)
			goto match;
	}

	for (i = 0; i < DSP_NUM_GPRS; i++) {
		/* assign a new free GPR */
		if (mgr->gpr[i].type == GPR_TYPE_NULL)
			goto match;
	}

	fprintf(stderr, "get_dynamic_gpr(): no available gprs\n");
	return -1;

      match:

	set_bit(i, used);
	mgr->gpr[i].type = GPR_TYPE_DYNAMIC;
	mgr->gpr[i].addr = i;
	mgr->gpr[i].usage++;

	return i + GPR_BASE;
}

static int get_static_gpr(struct dsp_patch_manager *mgr, __s32 val, __u32 * used)
{
	int i;

	/* these can't be shared so just get a new one */
	for (i = 0; i < DSP_NUM_GPRS; i++) {
		/* free */
		if (mgr->gpr[i].type == GPR_TYPE_NULL)
			goto match;
	}

	fprintf(stderr, "get_static_gpr(): no available gprs\n");
	return -1;

      match:

	set_bit(i, used);
	mgr->gpr[i].type = GPR_TYPE_STATIC;
	mgr->gpr[i].addr = i;
	mgr->gpr[i].usage++;
	mgr->gpr[i].value = val;

	return GPR_BASE + i;
}

/* We don't match IO gprs here, we just don't waste any (maximum of 2 per line) */
static int get_io_gpr(struct dsp_patch_manager *mgr, int line, int type, __u32 * used, __u32 * input)
{
	int i;

	line &= 0x3f;

	/* return an already allocated one (but not in use by this patch */
	if (mgr->io[line][0] >= 0) {
		if (!test_bit(mgr->io[line][0], used)) {
			i = mgr->io[line][0];
			goto match;
		}
	}

	if (mgr->io[line][1] >= 0) {
		if (!test_bit(mgr->io[line][1], used)) {
			i = mgr->io[line][1];
			goto match;
		}
	}

	/* get a free one */
	for (i = 0; i < DSP_NUM_GPRS; i++) {
		if (mgr->gpr[i].type == GPR_TYPE_NULL) {
			if (mgr->io[line][0] < 0)
				mgr->io[line][0] = i;
			else if (mgr->io[line][1] < 0)
				mgr->io[line][1] = i;
			else {
				fprintf(stderr, "get_io_gpr(): more than two io gprs per line\n");
				return -1;
			}
			goto match;
		}
	}

	fprintf(stderr, "get_io_gpr(): no available gprs\n");
	return -1;

      match:

	set_bit(i, used);
	if (type == GPR_INPUT)
		set_bit(i, input);

	mgr->gpr[i].type = GPR_TYPE_IO;
	mgr->gpr[i].addr = i;
	mgr->gpr[i].line = line;
	mgr->gpr[i].usage++;
	mgr->gpr[i].value = 0;

	return GPR_BASE + i;
}

/* FIXME */
/* allow for non sharing of control gprs between identical patches */
static int get_control_gpr(struct dsp_patch_manager *mgr, __s32 val, __s32 min, __s32 max,
			   const char *name, __u32 * used, const char *patch_name)
{
	int i;

	//if another instance of same patch is loaded, we'll use it's control gpr:
	if(patch_name!=NULL)
		if( (i=find_control_gpr_addr(mgr, patch_name, name)) != -1){
			mgr->gpr[i-GPR_BASE].usage++;
			set_bit(i-GPR_BASE, used);
			return  i;
		}

	for (i = 0; i < DSP_NUM_GPRS; i++) {
		/* free */
		if (mgr->gpr[i].type == GPR_TYPE_NULL)
			goto match;
	}

	fprintf(stderr, "get_control_gpr(): no available gprs\n");
	return -1;

      match:

	set_bit(i, used);
	mgr->gpr[i].type = GPR_TYPE_CONTROL;
	mgr->gpr[i].addr = i;
	mgr->gpr[i].value = val;
	mgr->gpr[i].min_val = min;
	mgr->gpr[i].max_val = max;
	mgr->gpr[i].usage++;
	strncpy(mgr->gpr[i].name, name, DSP_GPR_NAME_SIZE);
	mgr->gpr[i].name[DSP_GPR_NAME_SIZE - 1] = '\0';

	return GPR_BASE + i;
}

/* get an already allocated gpr by address */
static int get_gpr(struct dsp_patch_manager *mgr, int addr, __u32 * used)
{
	if (test_bit(addr, used))
		return GPR_BASE + addr;
	else {
		if (mgr->gpr[addr].type == GPR_TYPE_NULL)
			return -1;

		if (mgr->gpr[addr].usage == 1 && mgr->gpr[addr].type == GPR_TYPE_STATIC)
			return -1;

		if (mgr->gpr[addr].usage == 255)
			return -1;

		set_bit(addr, used);
		mgr->gpr[addr].usage++;
	}

	return GPR_BASE + addr;
}

static int free_gpr(struct dsp_patch_manager *mgr, int addr, __u32 * used)
{
	if (!test_bit(addr, used))
		return -1;

	clear_bit(addr, used);

	mgr->gpr[addr].usage--;

	if (mgr->gpr[addr].usage == 0) {
		if (mgr->gpr[addr].type == GPR_TYPE_IO) {
			if (mgr->io[mgr->gpr[addr].line][0] == addr)
				mgr->io[mgr->gpr[addr].line][0] = -1;
			else
				mgr->io[mgr->gpr[addr].line][1] = -1;
		}

		mgr->gpr[addr].type = GPR_TYPE_NULL;
	}

	return 0;
}

static void free_all_gprs(struct dsp_patch_manager *mgr, __u32 * used)
{
	int i;

	for (i = 0; i < DSP_NUM_GPRS; i++)
		free_gpr(mgr, i, used);
}

static int read_patch_code(struct dsp_patch *patch, FILE * fp)
{
	patch->code = (__u32 *) malloc(patch->code_size * sizeof(__u32));

	if (patch->code == NULL) {
		fprintf(stderr, "read_patch_code(): no memory available\n");
		return -1;
	}

	fread(patch->code, sizeof(__u32), patch->code_size, fp);

	return 0;
}

/* FIXME */
/* Still missing the ALIGN and CLEAR bits */
static inline void read_tram_lines(__u32 tramb_start, __u32 * traml_addr,
				   __u8 * traml_size, __u16 base, __u32 type, FILE * fp)
{
	int i;
	__u8 lines;
	__u8 data_addr;
	__u32 addr_val;

	fread(&lines, sizeof(__u8), 1, fp);
	for (i = 0; i < lines; i++) {
		fread(&data_addr, sizeof(__u8), 1, fp);

		addr_change[TRAML_IDATA_BASE - GPR_BASE + data_addr] = base + *traml_size;

		addr_change[TRAML_IADDR_BASE - GPR_BASE + data_addr] =
		    base + (TRAML_IADDR_BASE - TRAML_IDATA_BASE) + *traml_size;

		fread(&addr_val, sizeof(__u32), 1, fp);

		traml_addr[*traml_size] = ((tramb_start + addr_val) & 0xfffff) | (type << 20);
		(*traml_size)++;
	}
}

/* FIXME restore tramb_?start on failure */
static int read_tram_block(struct dsp_patch_manager *mgr, struct dsp_patch *patch, __u32 * traml_iaddr,
			   __u32 * traml_eaddr, FILE * fp)
{
	int i;
	__u8 blocks;
	__u32 size;
	__u8 tmp;

	/* tram blocks and lines */
	fread(&blocks, sizeof(__u8), 1, fp);
	for (i = 0; i < blocks; i++) {
		fread(&size, sizeof(__u32), 1, fp);
		if (size == 0) {
			fprintf(stderr, "tram block of 0 size\n");
			return -1;
		}

		if (size <= TRAMB_ISIZE - mgr->tramb_ifree_start) {
			tmp = patch->traml_isize;

			/* read lines */
			read_tram_lines(patch->tramb_isize,
					traml_iaddr, &patch->traml_isize, TRAML_IDATA_BASE, TRAML_TYPE_READ, fp);

			/* write lines */
			read_tram_lines(patch->tramb_isize,
					traml_iaddr, &patch->traml_isize, TRAML_IDATA_BASE, TRAML_TYPE_WRITE, fp);

			/* a block with no tram lines */
			if (tmp == patch->traml_isize) {
				fprintf(stderr, "tram block with no tram lines\n");
				return -1;
			}

			/* no free tram space */
			if (mgr->traml_ifree_start + patch->traml_isize > TRAML_ISIZE) {
				fprintf(stderr, "no free internal tram lines\n");
				return -1;
			}

			patch->tramb_isize += size;
			mgr->tramb_ifree_start += size;

		} else if (size <= TRAMB_ESIZE - mgr->tramb_efree_start) {
			tmp = patch->traml_esize;

			/* read lines */
			read_tram_lines(patch->tramb_esize, traml_eaddr,
					&patch->traml_esize, TRAML_EDATA_BASE, TRAML_TYPE_READ, fp);

			/* write lines */
			read_tram_lines(patch->tramb_esize, traml_eaddr,
					&patch->traml_esize, TRAML_EDATA_BASE, TRAML_TYPE_WRITE, fp);

			/* a block with no tram lines */
			if (tmp == patch->traml_esize) {
				fprintf(stderr, "tram block with no tram lines\n");
				return -1;
			}

			/* no free tram space */
			if (mgr->traml_efree_start + patch->traml_esize > TRAML_ESIZE) {
				fprintf(stderr, "no free external tram lines\n");
				return -1;
			}
			patch->tramb_esize += size;
			mgr->tramb_efree_start += size;

		} else {
			fprintf(stderr, "no free tram blocks\n");
			return -1;
		}
	}

	mgr->traml_ifree_start += patch->traml_isize;
	mgr->traml_efree_start += patch->traml_esize;

	return 0;
}

/* FIXME */
/* only single input patches supported */
static int read_patch_header(struct dsp_patch_manager *mgr, struct dsp_patch *patch, FILE * fp)
{
	__u8 count, addr;
	__u32 traml_iaddr[TRAML_ISIZE];
	__u32 traml_eaddr[TRAML_ESIZE];
	int ret, i;

	/* patch name */
	fread(patch->name, sizeof(char), DSP_PATCH_NAME_SIZE, fp);

	/* input/output gprs */
	fread(&count, sizeof(__u8), 1, fp);
	if (count > 1) {
		fprintf(stderr, "patches with several inputs/outputs not yet supported %d\n", count);
		return -1;
	}

	for (i = 0; i < count; i++) {
		fread(&addr, sizeof(__u8), 1, fp);

		addr_change[addr] = get_io_gpr(mgr, patch->line, GPR_INPUT, patch->gpr_used, patch->gpr_input);

		fread(&addr, sizeof(__u8), 1, fp);

		addr_change[addr] = get_io_gpr(mgr, patch->line, GPR_OUTPUT, patch->gpr_used, patch->gpr_input);
	}

	/* dynamic gprs */
	fread(&count, sizeof(__u8), 1, fp);
	for (i = 0; i < count; i++) {
		fread(&addr, sizeof(__u8), 1, fp);
		addr_change[addr] = get_dynamic_gpr(mgr, patch->gpr_used);
	}

	/* static gprs */
	fread(&count, sizeof(__u8), 1, fp);
	for (i = 0; i < count; i++) {
		__s32 value;
		fread(&addr, sizeof(__u8), 1, fp);
		fread(&value, sizeof(__s32), 1, fp);
		addr_change[addr] = get_static_gpr(mgr, value, patch->gpr_used);
	}

	/* control gprs */
	fread(&count, sizeof(__u8), 1, fp);
	for (i = 0; i < count; i++) {
		__s32 value, min, max;
		char name[DSP_GPR_NAME_SIZE];

		fread(&addr, sizeof(__u8), 1, fp);
		fread(&value, sizeof(__s32), 1, fp);
		fread(&min, sizeof(__s32), 1, fp);
		fread(&max, sizeof(__s32), 1, fp);

		if (min >= max) {
			fprintf(stderr, "invalid range for control gpr %x %x\n", min, max);
			return -1;
		}

		if (value > max)
			value = max;
		else if (value < min)
			value = min;

		fread(name, sizeof(char), DSP_GPR_NAME_SIZE, fp);

		addr_change[addr] = get_control_gpr(mgr, value, min, max, name, patch->gpr_used, patch->name);
	}

	/* constant gprs */
	fread(&count, sizeof(__u8), 1, fp);
	for (i = 0; i < count; i++) {
		__s32 value;

		fread(&addr, sizeof(__u8), 1, fp);
		fread(&value, sizeof(__s32), 1, fp);
		addr_change[addr] = get_constant_gpr(mgr, value, patch->gpr_used);
	}
	/* here lines and blocks, internal and external, all start from 0 */
	/* this is fixed later on in set_tram_addr() */

	patch->tramb_istart = 0;
	patch->tramb_isize = 0;
	patch->tramb_estart = 0;
	patch->tramb_esize = 0;

	patch->traml_istart = 0;
	patch->traml_isize = 0;

	patch->traml_estart = 0;
	patch->traml_esize = 0;

	/* tram tablelookup blocks and lines */
	ret = read_tram_block(mgr, patch, traml_iaddr, traml_eaddr, fp);
	if (ret < 0)
		return ret;

	/* tram delaylines blocks and lines */
	ret = read_tram_block(mgr, patch, traml_iaddr, traml_eaddr, fp);
	if (ret < 0)
		return ret;

	if (patch->traml_isize != 0) {
		patch->traml_iaddr = (__u32 *) malloc(patch->traml_isize * sizeof(__u32));
		if (patch->traml_iaddr == NULL) {
			fprintf(stderr, "read_patch_header(): no memory available\n");
			return -1;
		}

		memcpy(patch->traml_iaddr, traml_iaddr, patch->traml_isize * sizeof(__u32));
	}

	if (patch->traml_esize != 0) {
		patch->traml_eaddr = (__u32 *) malloc(patch->traml_esize * sizeof(__u32));
		if (patch->traml_eaddr == NULL) {
			fprintf(stderr, "read_patch_header(): no memory available\n");
			goto err0;
		}

		memcpy(patch->traml_eaddr, traml_eaddr, patch->traml_esize * sizeof(__u32));
	}

	fread(&patch->code_size, sizeof(__u16), 1, fp);

	if (patch->code_size == 0) {
		fprintf(stderr, "patch with zero code size\n");
		goto err1;
	}

	mgr->code_free_start += patch->code_size;

	if (mgr->code_free_start > DSP_CODE_SIZE) {
		fprintf(stderr, "no free dsp code memory\n");
		goto err2;
	}

	return 0;

      err2:
	mgr->code_free_start -= patch->code_size;

      err1:
	if (patch->traml_esize != 0)
		free(patch->traml_eaddr);

      err0:
	if (patch->traml_isize != 0)
		free(patch->traml_iaddr);

	return -1;
}

static void init_addr_change_table(__u32 * table, __u32 min, __u32 max)
{
	int i;

	for (i = min; i < max; i++)
		table[i - min] = i;

}

static void search_and_replace(__u32 * code, __u32 size, __u32 * table, __u32 min, __u32 max)
{
	__u32 operand[2];
	int i;

	for (i = 0; i < size; i++) {
		operand[0] = code[i] & 0x000003ff;
		operand[1] = (code[i] & 0x000ffc00) >> 10;

		if (operand[0] < max && operand[0] >= min)
			operand[0] = table[operand[0] - min];

		if (operand[1] < max && operand[1] >= min)
			operand[1] = table[operand[1] - min];

		code[i] = (code[i] & ~0x000fffff) | ((operand[1] & 0x3ff) << 10)
		    | (operand[0] & 0x3ff);
	}

	return;
}

/* gives an increasing id to patches with the same name */
static int set_patch_id(struct dsp_patch_manager *mgr, struct dsp_patch *patch)
{
	struct dsp_patch *patch1;
	struct list_head *entry;
	int table[255];
	int i;

	patch->id = 0;

	for (i = 0; i < DSP_NUM_INPUTS; i++)
		list_for_each(entry, &mgr->ipatch_list[i]) {
		patch1 = list_entry(entry, struct dsp_patch, list);

		if (!strcmp(patch1->name, patch->name)) {
			table[patch1->id] = 1;
			while (table[patch->id] == 1) {
				patch->id++;
				if (patch->id == 255)
					return -1;
			}
		}
		}

	for (i = 0; i < DSP_NUM_OUTPUTS; i++)
		list_for_each(entry, &mgr->opatch_list[i]) {
		patch1 = list_entry(entry, struct dsp_patch, list);

		if (!strcmp(patch1->name, patch->name)) {
			table[patch1->id] = 1;
			while (table[patch->id] == 1) {
				patch->id++;
				if (patch->id == 255)
					return -1;
			}
		}
		}

	return 0;
}

/* FIXME */
/* support patches with multiple inputs/outputs */
/* support sharing control gprs between identical patches */
/* support precise patch positioning in a line (patches in general don't commute) */
int dsp_read_patch(struct dsp_patch_manager *mgr, const char *file, const char *line)
{
	struct dsp_patch *patch;
	struct list_head *list;
	FILE *fp;
	int i, j;

	if (!mgr->init)
		dsp_init(mgr);

	patch = (struct dsp_patch *) malloc(sizeof(struct dsp_patch));
	if (patch == NULL) {
		perror("patch");
		goto err0;
	}
	memset(patch, 0, sizeof(struct dsp_patch));

	if ((fp = fopen(file, "r")) == NULL) {
		perror(file);
		goto err1;
	}

	for (i = 0; i < DSP_NUM_INPUTS; i++) {
		if (!strcmp(line, dsp_in_name[i])) {
			list = &mgr->ipatch_list[i];
			patch->line = i | (PATCH_INPUT << 6);

			/* only load a patch if there is a route connecting it */
			for (j = 0; j < DSP_NUM_OUTPUTS; j++)
				if (test_bit(i, &mgr->rpatch.route_v[j]) || test_bit(i, &mgr->rpatch.route[j]))
					goto match;

//                      set_bit(i, &patch->line);

		}
	}

	for (i = 0; i < DSP_NUM_OUTPUTS; i++) {
		if (!strcmp(line, dsp_out_name[i])) {
			list = &mgr->opatch_list[i];
			patch->line = i | (PATCH_OUTPUT << 6);

			/* only load a patch if there is a route connecting it */
			if (mgr->rpatch.route_v[i] || mgr->rpatch.route[i])
				goto match;

//                      set_bit(i, &patch->line);
		}
	}

      err2:
	fclose(fp);
      err1:
	free(patch);
      err0:
	return -1;

      match:

	if (read_patch_header(mgr, patch, fp) < 0)
		goto err2;

	read_patch_code(patch, fp);
	search_and_replace(patch->code, patch->code_size, addr_change, GPR_BASE, TRAML_EADDR_BASE + TRAML_ESIZE);

	set_patch_id(mgr, patch);
	determine_io(mgr, patch);

	list_add(&patch->list, list);

	fclose(fp);
	return 0;
}

int dsp_unload_patch(struct dsp_patch_manager *mgr, const char *name)
{
	struct dsp_patch *patch;
	struct list_head *entry;
	char s[DSP_PATCH_NAME_SIZE + 4];
	int i;

	if (!mgr->init)
		dsp_init(mgr);

	for (i = 0; i < DSP_NUM_INPUTS; i++)
		list_for_each(entry, &mgr->ipatch_list[i]) {
		patch = list_entry(entry, struct dsp_patch, list);
		if (patch->id)
			sprintf(s, "%s %d", patch->name, patch->id);
		else
			sprintf(s, "%s", patch->name);

		if (!strcmp(s, name))
			goto match;
		}


	for (i = 0; i < DSP_NUM_OUTPUTS; i++)
		list_for_each(entry, &mgr->opatch_list[i]) {
		patch = list_entry(entry, struct dsp_patch, list);

		if (patch->id)
			sprintf(s, "%s %d", patch->name, patch->id);
		else
			sprintf(s, "%s", patch->name);

		if (!strcmp(s, name))
			goto match;
		}

	return -1;

      match:
	list_del(&patch->list);
	free_all_gprs(mgr, patch->gpr_used);
	mgr->code_free_start -= patch->code_size;
	mgr->traml_efree_start -= patch->traml_esize;
	mgr->traml_ifree_start -= patch->traml_isize;
	mgr->tramb_efree_start -= patch->tramb_esize;
	mgr->tramb_ifree_start -= patch->tramb_isize;
	free(patch->traml_iaddr);
	free(patch->traml_eaddr);
	free(patch->code);
	free(patch);

	return 0;
}

int dsp_set_patch_name(struct dsp_patch_manager *mgr, const char *patch, const char *new)
{
	return -1;
}

int dsp_set_control_gpr_value(struct dsp_patch_manager *mgr, const char *patch, const char *gpr, __s32 val)
{
	struct mixer_private_ioctl ctl;
	int addr, fd, ret;

	if (mgr->init) {
		addr = find_control_gpr_addr(mgr, patch, gpr);
		if (addr < 0)
			return addr;

		addr -= GPR_BASE;

		if (val > mgr->gpr[addr].max_val)
			val = mgr->gpr[addr].max_val;
		else if (val < mgr->gpr[addr].min_val)
			val = mgr->gpr[addr].min_val;

		mgr->gpr[addr].value = val;
	}

	/* open device */
	if ((fd = open(MIXER_DEV_NAME, O_WRONLY, 0)) < 0) {
		perror(MIXER_DEV_NAME);
		return fd;
	}

	ctl.cmd = CMD_SETCTLGPR;
	memcpy((char *) ctl.val, patch, DSP_PATCH_NAME_SIZE);
	memcpy((char *) ctl.val + DSP_PATCH_NAME_SIZE, gpr, DSP_PATCH_NAME_SIZE);
	memcpy((char *) ctl.val + 2 * DSP_PATCH_NAME_SIZE, &val, sizeof(__s32));

	ret = ioctl(fd, SOUND_MIXER_PRIVATE3, &ctl);
	if (ret < 0) {
		perror("SOUND_MIXER_PRIVATE3");
		return ret;
	}

	return 0;
}

int dsp_get_control_gpr_value(struct dsp_patch_manager *mgr, const char *patch, const char *gpr, __s32 * val)
{
	struct mixer_private_ioctl ctl;
	int addr;
	int fd, ret;

	if (mgr->init) {
		addr = find_control_gpr_addr(mgr, patch, gpr);
		if (addr < 0)
			return addr;

		addr -= GPR_BASE;

		*val = mgr->gpr[addr].value;
	} else {
		/* open device */
		if ((fd = open(MIXER_DEV_NAME, O_RDONLY, 0)) < 0) {
			perror(MIXER_DEV_NAME);
			return fd;
		}

		ctl.cmd = CMD_GETCTLGPR;
		memcpy(ctl.val, patch, DSP_PATCH_NAME_SIZE);
		memcpy((char *) ctl.val + DSP_PATCH_NAME_SIZE, gpr, DSP_PATCH_NAME_SIZE);

		ret = ioctl(fd, SOUND_MIXER_PRIVATE3, &ctl);
		if (ret < 0) {
			perror("SOUND_MIXER_PRIVATE3");
			return ret;
		}

		*val = (__s32) ctl.val[0];
	}

	return 0;
}

static void construct_input_buffer(struct dsp_patch_manager *mgr, __u32 * operand, __u32 * buffer_code,
				   int *buffer_size)
{
	int addr;
	int i;

	for (i = 1; i < 3; i++)
		if (operand[i] < OUTPUT_BASE && operand[i + 1] < OUTPUT_BASE &&
		    addr_change[operand[i]] < OUTPUT_BASE && addr_change[operand[i + 1]] < OUTPUT_BASE) {
			addr = get_dynamic_gpr(mgr, mgr->rpatch.gpr_used);
			addr_change[operand[i]] = addr;
			buffer_code[*buffer_size + 1] = 0x6 << 20 | addr << 10 | operand[i];
			buffer_code[*buffer_size] = 0x40 << 10 | 0x40;
			*buffer_size += 2;
		}
}

static int construct_routing(struct dsp_patch_manager *mgr)
{
	struct dsp_rpatch *rpatch;
	struct list_head *entry;
	struct dsp_patch *patch;
	__u32 op, operand[4];
	char name[DSP_GPR_NAME_SIZE];
	__u32 buffer_code[DSP_NUM_INPUTS * 2];
	int buffer_size;
	int i, j;
	int route, route_v;
	int addr;

	rpatch = &mgr->rpatch;

	for (i = 0; i < DSP_NUM_INPUTS; i++)
		if (rpatch->in[i] >= GPR_BASE) {
			get_gpr(mgr, rpatch->in[i] - GPR_BASE, rpatch->gpr_used);
			set_bit(rpatch->in[i] - GPR_BASE, rpatch->gpr_input);
		}

	for (i = 0; i < DSP_NUM_OUTPUTS; i++)
		rpatch->out[i] = -1;

	buffer_size = 0;
	mgr->code_free_start -= rpatch->code_size;
	rpatch->code_size = 0;

	init_addr_change_table(addr_change, 0, OUTPUT_BASE);

	for (i = 0; i < DSP_NUM_OUTPUTS; i++) {
		if (mgr->rpatch.route[i] != 0 || mgr->rpatch.route_v[i] != 0) {
			route = 0;
			route_v = 0;
			j = 1;
			op = 0x6;

			while (route < DSP_NUM_INPUTS || route_v < DSP_NUM_INPUTS) {

				/* find next active route */
				while (!test_bit(route, &mgr->rpatch.route[i])
				       && route < DSP_NUM_INPUTS)
					route++;

				while (!test_bit(route_v, &mgr->rpatch.route_v[i])
				       && route_v < DSP_NUM_INPUTS)
					route_v++;

				/* a complete op and there are more arguments */
				if (j == 4 && (route < DSP_NUM_INPUTS || route_v < DSP_NUM_INPUTS)) {
					/* if we need an output gpr use it also for temporary storage */
					if (rpatch->out[i] < 0)
						rpatch->out[i] =
						    get_io_gpr(mgr, i, GPR_OUTPUT, rpatch->gpr_used, rpatch->gpr_input);

					if (rpatch->out[i] < 0)
						return -1;

					operand[0] = rpatch->out[i];

					rpatch->code[rpatch->code_size + 1] = op << 20 | operand[0] << 10 | operand[1];

					rpatch->code[rpatch->code_size] = operand[2] << 10 | operand[3];

					rpatch->code_size += 2;

					construct_input_buffer(mgr, operand, buffer_code, &buffer_size);

					op = 0x6;
					operand[1] = rpatch->out[i];
					j = 2;
				}

				if ((j == 2 || route >= DSP_NUM_INPUTS) && route_v < DSP_NUM_INPUTS) {
					if (j == 1)
						operand[1] = 0x40;

					operand[2] = rpatch->in[route_v];

					sprintf(name, "Vol %s:%s", dsp_in_name[route_v], dsp_out_name[i]);

					/* find control gpr for this route */
					/* if there isn't any create one */
					addr = find_control_gpr_addr(mgr, rpatch->name, name);
					if (addr < 0) {
						addr =
						    get_control_gpr(mgr, 0x7fffffff, 0, 0x7fffffff,
								    name, rpatch->gpr_used, NULL);
						if (addr < 0)
							return -1;
					}

					op = 0x0;
					operand[3] = addr;
					j = 4;
					route_v++;

				} else if (route < DSP_NUM_INPUTS) {
					operand[j++] = rpatch->in[route];
					route++;
				}
			}

			while (j < 4)
				operand[j++] = 0x40;

			addr = rpatch->out[i];
			rpatch->out[i] = OUTPUT_BASE + i;
			list_for_each(entry, &mgr->opatch_list[i]) {
				patch = list_entry(entry, struct dsp_patch, list);
				if (patch->in > 0) {
					if (addr < 0)
						rpatch->out[i] = get_io_gpr(mgr, i,
									    GPR_OUTPUT, rpatch->gpr_used,
									    rpatch->gpr_input);
					else
						rpatch->out[i] = addr;
					break;
				}
			}

			operand[0] = rpatch->out[i];

			rpatch->code[rpatch->code_size + 1] = op << 20 | operand[0] << 10 | operand[1];

			rpatch->code[rpatch->code_size] = operand[2] << 10 | operand[3];
			rpatch->code_size += 2;

			construct_input_buffer(mgr, operand, buffer_code, &buffer_size);
		}
	}

	search_and_replace(rpatch->code, rpatch->code_size, addr_change, 0, OUTPUT_BASE);

	if (buffer_size + rpatch->code_size > DSP_CODE_SIZE) {
		fprintf(stderr, "routing patch to big\n");
		return -1;
	}

	memmove(rpatch->code + buffer_size, rpatch->code, rpatch->code_size * sizeof(__u32));
	memcpy(rpatch->code, buffer_code, buffer_size * sizeof(__u32));
	rpatch->code_size += buffer_size;
	mgr->code_free_start += rpatch->code_size;

	if (mgr->code_free_start > DSP_CODE_SIZE) {
		fprintf(stderr, "no free dsp code memory\n");
		return -1;
	}

	return 0;
}


static int load_patch_code(__u32 * code, __u16 start, __u16 size, int fd)
{
	copr_buffer buf;
	int ret;

	buf.command = CMD_WRITE;
	buf.offs = DSP_CODE_BASE + start;
	buf.len = size < 0x200 ? size : 0x200;

	memcpy(buf.data, code, buf.len * sizeof(__u32));

	ret = ioctl(fd, SNDCTL_COPR_LOAD, &buf);
	if (ret < 0) {
		perror("SNDCTL_COPR_LOAD");
		return ret;
	}

	if (buf.len < size) {
		buf.offs = DSP_CODE_BASE + start + buf.len;
		buf.len = size - buf.len;

		memcpy(buf.data, &code[buf.offs - DSP_CODE_BASE], buf.len * sizeof(__u32));

		ret = ioctl(fd, SNDCTL_COPR_LOAD, &buf);
		if (ret < 0) {
			perror("SNDCTL_COPR_LOAD");
			return ret;
		}
	}

	return 0;
}

static int load_gprs(struct dsp_patch_manager *mgr, int audio_fd, int mixer_fd)
{
	copr_buffer buf;
	struct mixer_private_ioctl ctl;
	int i, ret;

	ctl.cmd = CMD_SETGPR;

	buf.command = CMD_WRITE;
	buf.len = 1;

	for (i = 0; i < DSP_NUM_GPRS; i++) {
		ctl.val[0] = i;

		memcpy(&ctl.val[1], &mgr->gpr[i], sizeof(struct dsp_kernel_gpr));

		ret = ioctl(mixer_fd, SOUND_MIXER_PRIVATE3, &ctl);
		if (ret < 0) {
			perror("SOUND_MIXER_PRIVATE3");
			return ret;
		}

		buf.offs = GPR_BASE + i;

		((__u32 *) buf.data)[0] = mgr->gpr[i].value;

		ret = ioctl(audio_fd, SNDCTL_COPR_LOAD, &buf);
		if (ret < 0) {
			perror("SNDCTL_COPR_LOAD");
			return ret;
		}
	}

	return 0;
}

static int load_patch_tramls(struct dsp_patch *patch, int fd)
{
	copr_buffer buf;
	int ret;

	buf.command = CMD_WRITE;
	buf.offs = TRAML_IADDR_BASE + patch->traml_istart;
	buf.len = patch->traml_isize;

	memcpy(buf.data, patch->traml_iaddr, buf.len * sizeof(__u32));

	ret = ioctl(fd, SNDCTL_COPR_LOAD, &buf);
	if (ret < 0) {
		perror("SNDCTL_COPR_LOAD");
		return ret;
	}

	buf.command = CMD_WRITE;
	buf.offs = TRAML_EADDR_BASE + patch->traml_estart;
	buf.len = patch->traml_esize;

	memcpy(buf.data, patch->traml_eaddr, buf.len * sizeof(__u32));

	ret = ioctl(fd, SNDCTL_COPR_LOAD, &buf);
	if (ret < 0) {
		perror("SNDCTL_COPR_LOAD");
		return ret;
	}

	return 0;
}

static void match_io(struct dsp_patch_manager *mgr, struct dsp_patch *patch, int *in)
{
	int gpr0, gpr1;

	gpr0 = mgr->io[patch->line & 0x3f][0];
	gpr1 = mgr->io[patch->line & 0x3f][1];

	if (patch->in != *in) {
		if (patch->in == -1 && patch->out == -1)
			return;

		/* match the input */
		addr_change[patch->in] = *in;

		if (patch->in >= GPR_BASE)
			clear_bit(patch->in - GPR_BASE, patch->gpr_input);

		if (*in >= GPR_BASE)
			set_bit(*in - GPR_BASE, patch->gpr_input);

		if (patch->out == *in) {
			/* the output and input are now swapped */
			if (patch->in < GPR_BASE) {
				/* the input was not an io gpr */
				if (*in == GPR_BASE + gpr1)
					patch->in = GPR_BASE + gpr0;
				else
					patch->in = GPR_BASE + gpr1;

				get_gpr(mgr, patch->in - GPR_BASE, patch->gpr_used);
			}
			addr_change[patch->out] = patch->in;
			*in = patch->in;
		} else {
			if (patch->in >= GPR_BASE)
				free_gpr(mgr, patch->in - GPR_BASE, patch->gpr_used);

			if (*in >= GPR_BASE)
				get_gpr(mgr, *in - GPR_BASE, patch->gpr_used);

			*in = patch->out;
		}
	} else {
		if (patch->out != -1)
			/* already matched */
			*in = patch->out;
	}
}

static void set_io_gprs(struct dsp_patch_manager *mgr)
{
	struct dsp_patch *patch;
	struct list_head *entry;
	int in, i;

	for (i = 0; i < DSP_NUM_INPUTS; i++) {
		in = i;

		list_for_each(entry, &mgr->ipatch_list[i]) {
			patch = list_entry(entry, struct dsp_patch, list);

			init_addr_change_table(addr_change, 0, GPR_BASE + GPR_SIZE);

			match_io(mgr, patch, &in);

			search_and_replace(patch->code, patch->code_size, addr_change, 0, GPR_BASE + GPR_SIZE);
		}
		mgr->rpatch.in[i] = in;
	}

	construct_routing(mgr);

	for (i = 0; i < DSP_NUM_OUTPUTS; i++) {
		in = mgr->rpatch.out[i];

		list_for_each(entry, &mgr->opatch_list[i]) {
			patch = list_entry(entry, struct dsp_patch, list);

			init_addr_change_table(addr_change, 0, GPR_BASE + GPR_SIZE);

			match_io(mgr, patch, &in);

			if (entry->next == &mgr->opatch_list[i] && patch->out > 0) {
				/* last patch in the line */
				addr_change[patch->out] = OUTPUT_BASE + i;
				if (patch->out >= GPR_BASE)
					free_gpr(mgr, patch->out - GPR_BASE, patch->gpr_used);
			}

			search_and_replace(patch->code, patch->code_size, addr_change, 0, GPR_BASE + GPR_SIZE);
		}
	}
}

static void set_code_start(struct dsp_patch_manager *mgr)
{
	struct dsp_patch *patch;
	struct list_head *entry;
	int code_start;
	int i;

	code_start = 0;

	for (i = 0; i < DSP_NUM_INPUTS; i++)
		list_for_each(entry, &mgr->ipatch_list[i]) {
		patch = list_entry(entry, struct dsp_patch, list);
		patch->code_start = code_start;
		code_start += patch->code_size;
		}

	mgr->rpatch.code_start = code_start;
	code_start += mgr->rpatch.code_size;

	for (i = 0; i < DSP_NUM_OUTPUTS; i++)
		list_for_each(entry, &mgr->opatch_list[i]) {
		patch = list_entry(entry, struct dsp_patch, list);
		patch->code_start = code_start;
		code_start += patch->code_size;
		}
}

/* a patch always have tramb_isize < tramb_esize when it gets here */
static void set_tram_addr(struct dsp_patch_manager *mgr, struct dsp_patch *patch)
{
	int j;

	if (patch->tramb_isize <= TRAMB_ISIZE - mgr->tramb_ifree_start &&
	    patch->traml_isize <= TRAML_ISIZE - mgr->traml_ifree_start) {

		/* keep internal block in internal tram */
		for (j = 0; j < patch->traml_isize; j++) {
			addr_change[patch->traml_istart + j] = TRAML_IDATA_BASE + mgr->traml_ifree_start + j;

			addr_change[TRAML_IADDR_BASE - TRAML_IDATA_BASE + patch->traml_istart + j] =
			    TRAML_IADDR_BASE + mgr->traml_ifree_start + j;

			patch->traml_iaddr[j] += mgr->tramb_ifree_start - patch->tramb_istart;
		}

		patch->tramb_istart = mgr->tramb_ifree_start;
		patch->traml_istart = mgr->traml_ifree_start;

		/* move external block to internal tram if possible */
		if (patch->tramb_esize <= TRAMB_ISIZE - (mgr->tramb_ifree_start + patch->tramb_isize)
		    && patch->traml_esize <=
		    TRAML_ISIZE - (mgr->traml_ifree_start + patch->traml_isize) && patch->tramb_esize != 0) {

//                      fprintf(stderr, "moving external block to internal tram\n");

			/* move external block to internal tram */
			patch->traml_iaddr = realloc(patch->traml_iaddr, patch->traml_isize + patch->traml_esize);

			for (j = 0; j < patch->traml_esize; j++) {
				patch->traml_iaddr[patch->traml_isize + j] =
				    patch->traml_eaddr[j] - patch->tramb_estart +
				    patch->tramb_istart + patch->tramb_isize;

				addr_change[TRAML_EDATA_BASE - TRAML_IDATA_BASE +
					    patch->traml_estart + j] =
				    TRAML_IDATA_BASE + mgr->traml_ifree_start + patch->traml_isize + j;

				addr_change[TRAML_EADDR_BASE - TRAML_IDATA_BASE +
					    patch->traml_estart + j] =
				    TRAML_EADDR_BASE + mgr->traml_ifree_start + patch->traml_isize + j;
			}

			patch->traml_isize += patch->traml_esize;
			patch->tramb_isize += patch->tramb_esize;
			mgr->tramb_ifree_start += patch->tramb_isize;
			mgr->traml_ifree_start += patch->traml_isize;
			patch->traml_esize = 0;
			patch->tramb_esize = 0;
			free(patch->traml_eaddr);
			patch->traml_eaddr = NULL;

		} else {
//                      fprintf(stderr, "keeping external block in external tram\n");

			mgr->tramb_ifree_start += patch->tramb_isize;
			mgr->traml_ifree_start += patch->traml_isize;

			/* keep external block in external tram */
			for (j = 0; j < patch->traml_esize; j++) {
				addr_change[TRAML_EDATA_BASE - TRAML_IDATA_BASE +
					    patch->traml_estart + j] = TRAML_EDATA_BASE + mgr->traml_efree_start + j;

				addr_change[TRAML_EADDR_BASE - TRAML_IDATA_BASE +
					    patch->traml_estart + j] = TRAML_EADDR_BASE + mgr->traml_efree_start + j;

				patch->traml_eaddr[j] += mgr->tramb_efree_start - patch->tramb_estart;
			}

			patch->tramb_estart = mgr->tramb_efree_start;
			mgr->tramb_efree_start += patch->tramb_esize;
			patch->traml_estart = mgr->traml_efree_start;
			mgr->traml_efree_start += patch->traml_esize;
		}

	} else {
		/* keep external block in external tram */
		for (j = 0; j < patch->traml_esize; j++) {
			addr_change[TRAML_EDATA_BASE - TRAML_IDATA_BASE + patch->traml_estart + j] =
			    TRAML_EDATA_BASE + mgr->traml_efree_start + j;

			addr_change[TRAML_EADDR_BASE - TRAML_IDATA_BASE + patch->traml_estart + j] =
			    TRAML_EADDR_BASE + mgr->traml_efree_start + j;

			patch->traml_eaddr[j] += mgr->tramb_efree_start - patch->tramb_estart;
		}

		patch->tramb_estart = mgr->tramb_efree_start;
		patch->traml_estart = mgr->traml_efree_start;

//              fprintf(stderr, "moving internal block to external tram\n");
		/* move internal block to external tram */
		patch->traml_eaddr = realloc(patch->traml_eaddr, patch->traml_isize + patch->traml_esize);

		for (j = 0; j < patch->traml_isize; j++) {
			patch->traml_eaddr[patch->traml_esize + j] =
			    patch->traml_iaddr[j] - patch->tramb_istart + patch->tramb_estart + patch->tramb_esize;

			addr_change[patch->traml_istart + j] =
			    TRAML_EDATA_BASE + mgr->traml_efree_start + patch->traml_esize + j;

			addr_change[TRAML_IADDR_BASE - TRAML_IDATA_BASE + patch->traml_istart + j] =
			    TRAML_EADDR_BASE + mgr->traml_efree_start + patch->traml_esize + j;
		}

		patch->traml_esize += patch->traml_isize;
		patch->tramb_esize += patch->tramb_isize;
		mgr->tramb_efree_start += patch->tramb_esize;
		mgr->traml_efree_start += patch->traml_esize;
		patch->traml_isize = 0;
		patch->tramb_isize = 0;
		free(patch->traml_iaddr);
		patch->traml_iaddr = NULL;

		return;
	}

	search_and_replace(patch->code, patch->code_size, addr_change, TRAML_IDATA_BASE,
			   TRAML_EADDR_BASE + TRAML_ESIZE);
}


static void set_tram(struct dsp_patch_manager *mgr)
{
	struct dsp_patch *patch;
	struct list_head *entry;
	int i;

	mgr->traml_ifree_start = 0;
	mgr->traml_efree_start = 0;

	mgr->tramb_ifree_start = 0;
	mgr->tramb_efree_start = 0;


	for (i = 0; i < DSP_NUM_INPUTS; i++)
		list_for_each(entry, &mgr->ipatch_list[i]) {
		patch = list_entry(entry, struct dsp_patch, list);
		set_tram_addr(mgr, patch);
		}

	for (i = 0; i < DSP_NUM_OUTPUTS; i++)
		list_for_each(entry, &mgr->opatch_list[i]) {
		patch = list_entry(entry, struct dsp_patch, list);
		set_tram_addr(mgr, patch);
		}
}

static int clean_free_dsp(struct dsp_patch_manager *mgr, int fd)
{
	copr_buffer buf;
	int i, size;
	int ret;

	size = mgr->icode_free_start - mgr->code_free_start;

	if (size <= 0)
		return 0;

	for (i = 0; i < (size < 1000 ? size : 1000) / 2; i++) {
		((__u32 *) buf.data)[2 * i] = 0x40 << 10 | 0x40;
		((__u32 *) buf.data)[2 * i + 1] = 6 << 20 | 0x40 << 10 | 0x40;
	}

	buf.command = CMD_WRITE;
	buf.offs = DSP_CODE_BASE + mgr->code_free_start;
	buf.len = size < 0x200 ? size : 0x200;

	ret = ioctl(fd, SNDCTL_COPR_LOAD, &buf);
	if (ret < 0) {
		perror("SNDCTL_COPR_LOAD");
		return ret;
	}

	if (buf.len < size) {
		buf.offs = DSP_CODE_BASE + mgr->code_free_start + buf.len;
		buf.len = size - buf.len;

		ret = ioctl(fd, SNDCTL_COPR_LOAD, &buf);
		if (ret < 0) {
			perror("SNDCTL_COPR_LOAD");
			return ret;
		}
	}

	return 0;
}


static int load_patches(struct dsp_patch_manager *mgr, int audio_fd, int mixer_fd)
{
	struct mixer_private_ioctl ctl;
	struct list_head *entry;
	struct dsp_patch *patch;
	struct dsp_rpatch *rpatch;
	int id, i, ret;

	set_io_gprs(mgr);
	set_code_start(mgr);
	set_tram(mgr);

	id = 0;

	ctl.cmd = CMD_SETPATCH;
	ctl.val[0] = id;

	/* routing patch */
	rpatch = &mgr->rpatch;

	memcpy(ctl.val + 1, rpatch, sizeof(struct dsp_kernel_rpatch));

	ret = ioctl(mixer_fd, SOUND_MIXER_PRIVATE3, &ctl);
	if (ret < 0) {
		perror("SOUND_MIXER_PRIVATE3");
		return ret;
	}

	load_patch_code(rpatch->code, rpatch->code_start, rpatch->code_size, audio_fd);


	/* FIXME */
	/* don't load a patch if it wasn't modified */

	/* input patches */

	for (i = 0; i < DSP_NUM_INPUTS; i++)
		list_for_each(entry, &mgr->ipatch_list[i]) {
		patch = list_entry(entry, struct dsp_patch, list);

		id++;
		ctl.val[0] = id;

		memcpy(ctl.val + 1, patch, sizeof(struct kdsp_patch));

		ret = ioctl(mixer_fd, SOUND_MIXER_PRIVATE3, &ctl);
		if (ret < 0) {
			perror("SOUND_MIXER_PRIVATE3");
			return ret;
		}

		ret = load_patch_code(patch->code, patch->code_start, patch->code_size, audio_fd);
		if (ret < 0)
			return ret;

		load_patch_tramls(patch, audio_fd);
		}

	/* output patches */

	for (i = 0; i < DSP_NUM_OUTPUTS; i++)
		list_for_each(entry, &mgr->opatch_list[i]) {
		patch = list_entry(entry, struct dsp_patch, list);

		id++;
		ctl.val[0] = id;

		memcpy(ctl.val + 1, patch, sizeof(struct kdsp_patch));

		ret = ioctl(mixer_fd, SOUND_MIXER_PRIVATE3, &ctl);
		if (ret < 0) {
			perror("SOUND_MIXER_PRIVATE3");
			return ret;
		}

		ret = load_patch_code(patch->code, patch->code_start, patch->code_size, audio_fd);
		if (ret < 0)
			return ret;

		load_patch_tramls(patch, audio_fd);
		}

	/* delimiter patch */

	id++;
	ctl.val[0] = id;

	((struct dsp_patch *) (ctl.val + 1))->code_size = 0;

	ret = ioctl(mixer_fd, SOUND_MIXER_PRIVATE3, &ctl);
	if (ret < 0) {
		perror("SOUND_MIXER_PRIVATE3");
		return ret;
	}

	clean_free_dsp(mgr, audio_fd);

	return 0;
}

static int set_external_tram(struct dsp_patch_manager *mgr, int fd)
{
	__u32 size;
	int ret;

	/* size here is in bytes */
	size = 2 * mgr->tramb_efree_start;

	if (!size)
		return 0;

	ret = ioctl(fd, SOUND_MIXER_PRIVATE4, &size);
	if (ret < 0) {
		perror("SOUND_MIXER_PRIVATE4");
		return ret;
	}

	return 0;
}


int dsp_load(struct dsp_patch_manager *mgr)
{
	int audio_fd, mixer_fd;
	int i;
	int ret;

	if (!mgr->init)
		return -1;

	/* open device */
	if ((mixer_fd = open(MIXER_DEV_NAME, O_RDONLY, 0)) < 0) {
		perror(MIXER_DEV_NAME);
		ret = mixer_fd;
		goto err0;
	}

	if ((audio_fd = open(DSP_DEV_NAME, O_RDONLY, 0)) < 0) {
		perror(DSP_DEV_NAME);
		ret = audio_fd;
		goto err1;
	}

	ret = set_external_tram(mgr, mixer_fd);
	if (ret < 0)
		goto err2;

	ret = load_patches(mgr, audio_fd, mixer_fd);
	if (ret < 0)
		goto err2;

	ret = load_gprs(mgr, audio_fd, mixer_fd);
	if (ret < 0)
		goto err2;

	if (mgr->debug)
		print_debug(mgr);

	/* release io and dynamic gprs of the routing patch */
	/* this is done here so that we can call dsp_load() several times */
	for (i = 0; i < DSP_NUM_GPRS; i++)
		if (test_bit(i, mgr->rpatch.gpr_used) &&
		    (mgr->gpr[i].type == GPR_TYPE_IO ||
		     mgr->gpr[i].type == GPR_TYPE_DYNAMIC)) free_gpr(mgr, i, mgr->rpatch.gpr_used);


	close(audio_fd);
	close(mixer_fd);

	return 0;

      err2:
	close(audio_fd);
      err1:
	close(mixer_fd);
      err0:
	return ret;
}


static int get_patch_code(struct dsp_patch *patch, int fd)
{
	copr_buffer buf;
	int ret;

	patch->code = (__u32 *) malloc(patch->code_size * sizeof(__u32));
	if (patch->code == NULL) {
		perror("patch code");
		return -1;
	}

	buf.command = CMD_READ;
	buf.offs = DSP_CODE_BASE + patch->code_start;
	buf.len = patch->code_size < 0x200 ? patch->code_size : 0x200;

	ret = ioctl(fd, SNDCTL_COPR_LOAD, &buf);
	if (ret < 0) {
		perror("SNDCTL_COPR_LOAD");
		return ret;
	}

	memcpy(patch->code, buf.data, buf.len * sizeof(__u32));

	if (patch->code_size > buf.len) {
		buf.offs = DSP_CODE_BASE + patch->code_start + buf.len;
		buf.len = patch->code_size - buf.len;

		ret = ioctl(fd, SNDCTL_COPR_LOAD, &buf);
		if (ret < 0) {
			perror("SNDCTL_COPR_LOAD");
			return ret;
		}

		memcpy(&(patch->code)[buf.offs - DSP_CODE_BASE], buf.data, buf.len * sizeof(__u32));
	}

	return 0;
}

static int get_patch_tramls(struct dsp_patch *patch, int fd)
{
	copr_buffer buf;
	int ret;

	patch->traml_iaddr = (__u32 *) malloc(patch->traml_isize * sizeof(__u32));

	if (patch->traml_iaddr == NULL) {
		perror("tram internal lines");
		return -1;
	}

	buf.command = CMD_READ;
	buf.offs = TRAML_IADDR_BASE + patch->traml_istart;
	buf.len = patch->traml_isize;

	ret = ioctl(fd, SNDCTL_COPR_LOAD, &buf);
	if (ret < 0) {
		perror("SNDCTL_COPR_LOAD");
		return ret;
	}

	memcpy(patch->traml_iaddr, buf.data, buf.len * sizeof(__u32));

	patch->traml_eaddr = (__u32 *) malloc(patch->traml_esize * sizeof(__u32));

	if (patch->traml_eaddr == NULL) {
		perror("tram external lines");
		return -1;
	}

	buf.command = CMD_READ;
	buf.offs = TRAML_EADDR_BASE + patch->traml_estart;
	buf.len = patch->traml_esize;

	ret = ioctl(fd, SNDCTL_COPR_LOAD, &buf);
	if (ret < 0) {
		perror("SNDCTL_COPR_LOAD");
		return ret;
	}

	memcpy(patch->traml_eaddr, buf.data, buf.len * sizeof(__u32));

	return 0;
}

static int get_gprs(struct dsp_patch_manager *mgr, int audio_fd, int mixer_fd)
{
	copr_buffer buf;
	struct mixer_private_ioctl ctl;
	int i;

	for (i = 0; i < DSP_NUM_GPRS; i++) {
		ctl.cmd = CMD_GETGPR;
		ctl.val[0] = i;

		if (ioctl(mixer_fd, SOUND_MIXER_PRIVATE3, &ctl) < 0) {
			perror("SOUND_MIXER_PRIVATE3");
			return -1;
		}

		memcpy(&mgr->gpr[i], &ctl, sizeof(struct dsp_kernel_gpr));

		buf.command = CMD_READ;
		buf.len = 1;
		buf.offs = GPR_BASE + i;

		if (ioctl(audio_fd, SNDCTL_COPR_LOAD, &buf) < 0) {
			perror("SNDCTL_COPR_LOAD");
			return -1;
		}

		mgr->gpr[i].value = ((__u32 *) buf.data)[0];
		mgr->gpr[i].addr = i;
	}

	return 0;
}


static int get_patches(struct dsp_patch_manager *mgr, int audio_fd, int mixer_fd)
{
	struct mixer_private_ioctl ctl;
	struct dsp_patch *patch;
	struct dsp_rpatch *rpatch;
	int i;

	/* input/ouput line list initialization */
	for (i = 0; i < DSP_NUM_INPUTS; i++)
		INIT_LIST_HEAD(&mgr->ipatch_list[i]);

	for (i = 0; i < DSP_NUM_OUTPUTS; i++)
		INIT_LIST_HEAD(&mgr->opatch_list[i]);

	mgr->traml_efree_start = 0;
	mgr->traml_ifree_start = 0;

	mgr->tramb_efree_start = 0;
	mgr->tramb_ifree_start = 0;

	mgr->code_free_start = 0;

	ctl.cmd = CMD_GETPATCH;

	/* get routing patch, id is always 0 */
	ctl.val[0] = 0;

	if (ioctl(mixer_fd, SOUND_MIXER_PRIVATE3, &ctl) < 0) {
		perror("SOUND_MIXER_PRIVATE3");
		return -1;
	}

	rpatch = &mgr->rpatch;

	memcpy(rpatch, &ctl, sizeof(struct dsp_kernel_rpatch));

	mgr->code_free_start += rpatch->code_size;

	/* always allocate extra space allowing the routing patch to grow later */
	rpatch->code = (__u32 *) malloc(DSP_CODE_SIZE * sizeof(__u32));
	if (rpatch->code == NULL) {
		perror("routing patch code");
		return -1;
	}

	/* get all other paches, stop when the code size is 0 */
	for (i = 1;; i++) {
		ctl.cmd = CMD_GETPATCH;
		ctl.val[0] = i;

		if (ioctl(mixer_fd, SOUND_MIXER_PRIVATE3, &ctl) < 0) {
			perror("SOUND_MIXER_PRIVATE3");
			return -1;
		}

		patch = (struct dsp_patch *) malloc(sizeof(struct dsp_patch));

		if (patch == NULL) {
			perror("patch");
			return -1;
		}

		memcpy(patch, &ctl, sizeof(struct kdsp_patch));

		if (patch->code_size == 0) {
			free(patch);
			break;
		}

		if (get_patch_code(patch, audio_fd) < 0)
			return -1;

		if (get_patch_tramls(patch, audio_fd) < 0)
			return -1;

		if ((patch->line >> 6) == PATCH_INPUT)
			list_add_tail(&patch->list, &mgr->ipatch_list[patch->line & 0x3f]);
		else {
			list_add_tail(&patch->list, &mgr->opatch_list[patch->line & 0x3f]);

		}

		mgr->code_free_start += patch->code_size;

		mgr->traml_ifree_start += patch->traml_isize;
		mgr->traml_efree_start += patch->traml_esize;

		mgr->tramb_ifree_start += patch->tramb_isize;
		mgr->tramb_efree_start += patch->tramb_esize;

	}

	return 0;
}

/* returns 1 if the code contains a particular address (val) */
static int search_code(struct dsp_patch *patch, int val)
{
	int i;

	for (i = 0; i < patch->code_size / 2; i++)
		if (((patch->code[i * 2 + 1] >> 10) & 0x3ff) == val ||
		    (patch->code[i * 2 + 1] & 0x3ff) == val ||
		    ((patch->code[i * 2] >> 10) & 0x3ff) == val || (patch->code[i * 2] & 0x3ff) == val)
			return 1;

	return 0;
}

static void determine_io(struct dsp_patch_manager *mgr, struct dsp_patch *patch)
{
	int gpr0, gpr1, line;

	line = patch->line & 0x3f;

	gpr0 = mgr->io[line][0];
	gpr1 = mgr->io[line][1];

	/* determine patch input */
	if (gpr0 != -1 && test_bit(gpr0, patch->gpr_input))
		patch->in = GPR_BASE + gpr0;
	else if (gpr1 != -1 && test_bit(gpr1, patch->gpr_input))
		patch->in = GPR_BASE + gpr1;
	else if (search_code(patch, line))
		patch->in = line;
	else
		patch->in = -1;

	/* determine patch output */
	if (gpr0 != -1 && test_bit(gpr0, patch->gpr_used) && !test_bit(gpr0, patch->gpr_input))
		patch->out = GPR_BASE + gpr0;
	else if (gpr1 != -1 && test_bit(gpr0, patch->gpr_used) && !test_bit(gpr1, patch->gpr_input))
		patch->out = GPR_BASE + gpr1;
	else if (search_code(patch, OUTPUT_BASE + line))
		patch->out = OUTPUT_BASE + line;
	else
		patch->out = -1;

}

static void determine_io_gprs(struct dsp_patch_manager *mgr)
{
	struct dsp_patch *patch;
	struct list_head *entry;
	int i;

	for (i = 0; i < DSP_NUM_INPUTS; i++)
		list_for_each(entry, &mgr->ipatch_list[i]) {
		patch = list_entry(entry, struct dsp_patch, list);
		determine_io(mgr, patch);
		}

	for (i = 0; i < DSP_NUM_OUTPUTS; i++)
		list_for_each(entry, &mgr->opatch_list[i]) {
		patch = list_entry(entry, struct dsp_patch, list);
		determine_io(mgr, patch);
		}
}

static void init_io_gprs_table(struct dsp_patch_manager *mgr)
{
	int line, i;

	for (i = 0; i < DSP_NUM_INPUTS; i++) {
		mgr->io[i][0] = -1;
		mgr->io[i][1] = -1;
	}

	/* construct table of possible io gprs for each line */
	for (i = 0; i < DSP_NUM_GPRS; i++)
		if (mgr->gpr[i].type == GPR_TYPE_IO) {
			line = mgr->gpr[i].line & 0x3f;
			if (mgr->io[line][0] == -1)
				mgr->io[line][0] = mgr->gpr[i].addr;
			else if (mgr->io[line][1] == -1)
				mgr->io[line][1] = mgr->gpr[i].addr;
			else {
				fprintf(stderr, "init_io_gprs_table(): more than 2 io gprs\n");
				exit(EXIT_FAILURE);
			}
		}
}

/* get all the dsp code information from the driver */
int dsp_init(struct dsp_patch_manager *mgr)
{
	int mixer_fd;
	int audio_fd;
	int ret;
	int i;

	/* open device */
	if ((mixer_fd = open(MIXER_DEV_NAME, O_RDONLY, 0)) < 0) {
		perror(MIXER_DEV_NAME);
		ret = mixer_fd;
		goto err0;
	}

	if ((audio_fd = open(DSP_DEV_NAME, O_RDONLY, 0)) < 0) {
		perror(DSP_DEV_NAME);
		ret = audio_fd;
		goto err1;
	}

	ret = get_patches(mgr, audio_fd, mixer_fd);
	if (ret < 0)
		goto err2;

	mgr->icode_free_start = mgr->code_free_start;

	ret = get_gprs(mgr, audio_fd, mixer_fd);
	if (ret < 0)
		goto err2;

	init_io_gprs_table(mgr);
	determine_io_gprs(mgr);

	/* release io and dynamic gprs of the routing patch */
	for (i = 0; i < DSP_NUM_GPRS; i++)
		if (test_bit(i, mgr->rpatch.gpr_used) &&
		    (mgr->gpr[i].type == GPR_TYPE_IO ||
		     mgr->gpr[i].type == GPR_TYPE_DYNAMIC)) free_gpr(mgr, i, mgr->rpatch.gpr_used);

	close(mixer_fd);
	close(audio_fd);

	mgr->init = 1;

	return 0;

      err2:
	close(audio_fd);

      err1:
	close(mixer_fd);

      err0:
	exit(EXIT_FAILURE);
}

void dsp_set_debug(struct dsp_patch_manager *mgr)
{
	if (!mgr->init)
		dsp_init(mgr);

	mgr->debug = 1;
}

static int get_input_name(const char *input)
{
	int i;

	for (i = 0; i < DSP_NUM_INPUTS; i++)
		if (!strcmp(input, dsp_in_name[i]))
			return i;

	return -1;
}

static int get_output_name(const char *output)
{
	int i;

	for (i = 0; i < DSP_NUM_OUTPUTS; i++)
		if (!strcmp(output, dsp_out_name[i]))
			return i;

	return -1;
}

static int get_route_name(struct dsp_patch_manager *mgr, const char *name, int *in, int *out)
{
	char route_name[2 * DSP_LINE_NAME_SIZE + 2];

	for (*in = 0; *in < DSP_NUM_INPUTS; (*in)++)
		for (*out = 0; *out < DSP_NUM_OUTPUTS; (*out)++) {
			sprintf(route_name, "%s:%s", dsp_in_name[*in], dsp_out_name[*out]);
			if (!strcmp(name, route_name))
				return 0;
		}

	return -1;
}

int dsp_check_input(struct dsp_patch_manager *mgr, int in)
{
	struct dsp_rpatch *rpatch = &mgr->rpatch;
	int i;

	if (in < 0 || in > DSP_NUM_INPUTS)
		return 0;

	for (i = 0; i < DSP_NUM_OUTPUTS; i++)
		if (test_bit(in, &rpatch->route[i]) || test_bit(in, &rpatch->route_v[i]))
			return 1;

	return 0;
}

int dsp_check_input_name(struct dsp_patch_manager *mgr, const char *input)
{
	int in;

	in = get_input_name(input);
	return dsp_check_input(mgr, in);
}

int dsp_check_output(struct dsp_patch_manager *mgr, int out)
{
	struct dsp_rpatch *rpatch = &mgr->rpatch;

	if (out < 0 || out > DSP_NUM_OUTPUTS)
		return 0;

	if (rpatch->route[out] || rpatch->route_v[out])
		return 1;

	return 0;
}

int dsp_check_output_name(struct dsp_patch_manager *mgr, const char *output)
{
	int out;

	out = get_output_name(output);
	return dsp_check_output(mgr, out);
}

int dsp_check_route(struct dsp_patch_manager *mgr, int in, int out)
{
	struct dsp_rpatch *rpatch = &mgr->rpatch;

	if (in < 0 || in > DSP_NUM_INPUTS || out < 0 || out > DSP_NUM_OUTPUTS)
		return 0;

	if (test_bit(in, &rpatch->route[out]) || test_bit(in, &rpatch->route_v[out]))
		return 1;

	return 0;
}

int dsp_check_route_volume(struct dsp_patch_manager *mgr, int in, int out)
{
	struct dsp_rpatch *rpatch = &mgr->rpatch;

	if (in < 0 || in > DSP_NUM_INPUTS || out < 0 || out > DSP_NUM_OUTPUTS)
		return 0;

	if (test_bit(in, &rpatch->route_v[out]))
		return 1;

	return 0;
}

int dsp_check_route_name(struct dsp_patch_manager *mgr, const char *name)
{
	int in, out;

	get_route_name(mgr, name, &in, &out);
	return dsp_check_route(mgr, in, out);
}

int dsp_del_route(struct dsp_patch_manager *mgr, int in, int out)
{
	struct dsp_rpatch *rpatch = &mgr->rpatch;
	char s[2 * DSP_LINE_NAME_SIZE + 5];
	int addr;
	int i;

	if (in < 0 || in > DSP_NUM_INPUTS || out < 0 || out > DSP_NUM_OUTPUTS)
		return -1;

	if (!mgr->init)
		dsp_init(mgr);

	if (test_and_clear_bit(in, &rpatch->route[out])) {
		/* don't remove the route if it is the last one connected to a dsp program */
		if (rpatch->route[out] || rpatch->route_v[out] || list_empty(&mgr->opatch_list[out])) {

			if (list_empty(&mgr->ipatch_list[in]))
				return 0;

			for (i = 0; i < DSP_NUM_OUTPUTS; i++)
				if (test_bit(in, &rpatch->route[i]) || test_bit(in, &rpatch->route_v[i]))
					return 0;
		}

		set_bit(in, &rpatch->route[out]);

	} else if (test_and_clear_bit(in, &rpatch->route_v[out])) {
		sprintf(s, "Vol %s:%s", dsp_in_name[in], dsp_out_name[out]);
		addr = find_control_gpr_addr(mgr, rpatch->name, s) - GPR_BASE;
		if (addr < 0)
			return -1;

		if (rpatch->route_v[out] || rpatch->route[out] || list_empty(&mgr->opatch_list[out])) {

			if (list_empty(&mgr->ipatch_list[in])) {
				free_gpr(mgr, addr, rpatch->gpr_used);
				return 0;
			}

			for (i = 0; i < DSP_NUM_OUTPUTS; i++)
				if (test_bit(in, &rpatch->route[i]) || test_bit(in, &rpatch->route_v[i])) {
					free_gpr(mgr, addr, rpatch->gpr_used);
					return 0;
				}
		}

		set_bit(in, &rpatch->route_v[out]);
	}

	return -1;
}

int dsp_del_route_name(struct dsp_patch_manager *mgr, const char *name)
{
	int in, out;

	get_route_name(mgr, name, &in, &out);
	return dsp_del_route(mgr, in, out);
}

int dsp_add_route(struct dsp_patch_manager *mgr, int in, int out)
{
	struct dsp_rpatch *rpatch = &mgr->rpatch;
	char s[2 * DSP_LINE_NAME_SIZE + 5];
	int addr;

	if (in < 0 || in > DSP_NUM_INPUTS || out < 0 || out > DSP_NUM_OUTPUTS)
		return -1;

	if (!mgr->init)
		dsp_init(mgr);

	set_bit(in, &rpatch->route[out]);

	if (test_bit(in, &rpatch->route_v[out])) {
		sprintf(s, "Vol %s:%s", dsp_in_name[in], dsp_out_name[out]);
		addr = find_control_gpr_addr(mgr, rpatch->name, s) - GPR_BASE;
		if (addr < 0)
			return -1;

		free_gpr(mgr, addr, rpatch->gpr_used);
		clear_bit(in, &rpatch->route_v[out]);
	}

	return 0;
}

int dsp_add_route_name(struct dsp_patch_manager *mgr, const char *name)
{
	int in, out;

	get_route_name(mgr, name, &in, &out);
	return dsp_add_route(mgr, in, out);
}

int dsp_add_route_v(struct dsp_patch_manager *mgr, int in, int out)
{
	char gpr_name[DSP_GPR_NAME_SIZE];

	if (in < 0 || in > DSP_NUM_INPUTS || out < 0 || out > DSP_NUM_OUTPUTS)
		return -1;

	if (!mgr->init)
		dsp_init(mgr);

	if (!test_and_set_bit(in, &mgr->rpatch.route_v[out])) {
		sprintf(gpr_name, "Vol %s:%s", dsp_in_name[in], dsp_out_name[out]);
		if (get_control_gpr(mgr, 0x7fffffff, 0, 0x7fffffff, gpr_name, mgr->rpatch.gpr_used, NULL) < 0)
			return -1;
	}

	clear_bit(in, &mgr->rpatch.route[out]);
	return 0;
}

int dsp_add_route_v_name(struct dsp_patch_manager *mgr, const char *name)
{
	int in, out;

	get_route_name(mgr, name, &in, &out);
	return dsp_add_route_v(mgr, in, out);
}

void dsp_print_inputs_name(void)
{
	int i;

	printf("Inputs:\n");

	for (i = 0; i < DSP_NUM_INPUTS; i++)
		printf("  %s\n", dsp_in_name[i]);
}
void dsp_print_outputs_name(void)
{
	int i;

	printf("Outputs:\n");

	for (i = 0; i < DSP_NUM_OUTPUTS; i++)
		printf("  %s\n", dsp_out_name[i]);
}
