/*********************************************************************
 *     mgr_text.c - text interface for the dsp program manager 
 * 
 *      Copyright (C) 2000 Rui Sousa 
 ********************************************************************* 
 *     This program is free software; you can redistribute it and/or 
 *     modify it under the terms of the GNU General Public License as 
 *     published by the Free Software Foundation; either version 2 of 
 *     the License, or (at your option) any later version. 
 * 
 *     This program is distributed in the hope that it will be useful, 
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of 
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 *     GNU General Public License for more details. 
 * 
 *     You should have received a copy of the GNU General Public 
 *     License along with this program; if not, write to the Free 
 *     Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, 
 *     USA. 
 ********************************************************************* 
*/


#include <argp.h>
#include "include/dsp.h"

const char *argp_program_version = "mgr_text 0.4";
const char *argp_program_bug_address = "<emu10k1-tools-devel@opensource.creative.com>";
const char doc[] = "mgr_text - text interface for the dsp microcode patch manager for emu10k1 based sound cards";

const struct argp_option options[] = {

	{"add", 'a', "ROUTE", 0,
	 "adds a route"},
	{"control", 'c', "CONTROL", OPTION_ARG_OPTIONAL,
	 "prints control gprs/selects control gpr to be modified"},
	{"debug", 'd', 0, 0, "enables debug output"},
	{"file", 'f', "FILE", 0, "loads a dsp microcode program"},
	{"inputs", 'i', 0, 0, "prints all the available inputs"},
	{"line", 'l', "LINE", 0,
	 "specifies the line to attatch the program"},
	{"name", 'n', "NAME", 0, "changes name of the selected patch"},
	{"outputs", 'o', 0, 0, "prints all the available outputs"},

	{"patch", 'p', "PATCH", 0,
	 "selects patch to change name or control gpr"},
	{"remove", 'r', "ROUTE", OPTION_ARG_OPTIONAL,
	 "prints existing routes/removes a route"},
	{"set", 's', "VALUE", 0,
	 "specifies a new value for the selected control gpr"},
	{"unload", 'u', "NAME", OPTION_ARG_OPTIONAL,
	 "lists loaded dsp programs/unloads a dsp microcode program"},
	{"volume", 'v', "ROUTE", 0,
	 "adds a route with volume control"},
	{0}
};

static error_t parser(int key, char *arg, struct argp_state *state)
{
	struct dsp_patch_manager *mgr = state->input;
	static char gpr[DSP_GPR_NAME_SIZE];
	static char patch[DSP_PATCH_NAME_SIZE];
	static char line[10];
	__s32 value;

	switch (key) {
	case 'a':
		dsp_add_route_name(mgr, arg);
		break;

	case 'c':
		if (arg == NULL)
			dsp_print_control_gpr_list(mgr);
		else {
			strncpy(gpr, arg, DSP_GPR_NAME_SIZE);
			gpr[DSP_GPR_NAME_SIZE - 1] = '\0';
		}
		break;

	case 'd':
		dsp_set_debug(mgr);
		break;
	case 'f':
		dsp_read_patch(mgr, arg, line);
		break;

	case 'i':
		dsp_print_inputs_name();
		break;

	case 'l':
		strncpy(line, arg, DSP_LINE_NAME_SIZE);
		line[DSP_LINE_NAME_SIZE - 1] = '\0';
		break;

	case 'n':
		dsp_set_patch_name(mgr, patch, arg);
		break;

	case 'o':
		dsp_print_outputs_name();
		break;

	case 'p':
		strncpy(patch, arg, DSP_PATCH_NAME_SIZE);
		patch[DSP_PATCH_NAME_SIZE - 1] = '\0';
		break;

	case 'r':
		if (arg == NULL)
			dsp_print_route_list(mgr);
		else
			dsp_del_route_name(mgr, arg);
		break;

	case 's':
		sscanf(arg, "%x", &value);
		dsp_set_control_gpr_value(mgr, patch, gpr, value);
		break;

	case 'u':
		if (arg == NULL)
			dsp_print_patch_list(mgr);
		else
			dsp_unload_patch(mgr, arg);

		break;
	case 'v':
		dsp_add_route_v_name(mgr, arg);
		break;
	default:
		return ARGP_ERR_UNKNOWN;

	}
	return 0;
}

static struct argp argp = { options, parser, 0, doc, };


int main(int argc, char **argv)
{
	struct dsp_patch_manager mgr;
	int ret;

	mgr.init = 0;
	mgr.debug = 0;

	argp_parse(&argp, argc, argv, 0, 0, &mgr);

	ret = dsp_load(&mgr);
	if (ret < 0)
		return ret;

	return 0;
}
